/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2008 OpenCFD Ltd.
      \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "colloidParticleCloud.H"

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool Foam::colloidParticle::move(colloidParticle::trackData& td)
{
    td.switchProcessor = false;
    td.keepParticle = true;

    const polyMesh& mesh = cloud().pMesh();
    const polyBoundaryMesh& pbMesh = mesh.boundaryMesh();

    scalar pi = mathematicalConstant::pi;

    scalar deltaT = mesh.time().deltaT().value();
    scalar tEnd = (1.0 - stepFraction())*deltaT;
    scalar dtMax = tEnd;

    //int counter = 0;

    while (td.keepParticle && !td.switchProcessor && tEnd > SMALL)
    {
        //counter++;
        //Info << "Trip through this loop: " << counter << endl;

        if (debug)
        {
            Info<< "Time = " << mesh.time().timeName()
                << " deltaT = " << deltaT
                << " tEnd = " << tEnd
                << " steptFraction() = " << stepFraction() << endl;
        }

        // set the lagrangian time-step
        scalar dt = min(dtMax, tEnd);

        // remember which cell the parcel is in
        // since this will change if a face is hit
        label celli = cell();

        //Info << "Moving particle at " << position_ << endl;
        dt *= trackToFace(position() + dt*U_, td);

        tEnd -= dt;
        stepFraction() = 1.0 - tEnd/deltaT;

        cellPointWeight cpw(mesh, position(), celli, face());
        scalar rhoc = td.spc().rhoc();
        vector Uc = td.UInterp().interpolate(cpw);
        scalar nuc = td.spc().nuc();

        scalar rhop = td.spc().rhop();
        scalar magUr = mag(Uc - U_);
        m_ = pi*d_*d_*d_/6.*rhop;

        // calculate drag effects
        scalar ReFunc = 1.0;
        scalar Re = magUr*d_/nuc;

        if (Re > 0.01)
        {
            ReFunc += 0.15*pow(Re, 0.687);
        }

        scalar Dc = (24.0*nuc/d_)*ReFunc*(3.0/4.0)*(rhoc/(d_*rhop));

        // Find distance between particle and nearest wall
        // and unit normal of that wall
        scalarField hDistance = scalarField(2,100.);
        vector hPointNew = vector(0.,0.,0.);
        vector hPointOld = vector(0.,0.,0.);
        label newh=1;
        label oldh =0;
        scalar h=0.;
        vector hPoint;
        vector Sf;
        vector Ucolloid = vector(0.,0.,0.);

        const polyBoundaryMesh& patches = td.spc().pMesh().boundaryMesh();

        //Info << "Finding wall distance..." << endl;

        forAll(patches,patchI)
        {
            const polyPatch& pPatch = patches[patchI];
            const pointField& points = pPatch.points();

            if(pPatch.type() == "wall")
            {
                forAll(pPatch,faceI)
                {
                    pointHit whoo = pPatch[faceI].nearestPoint(position(),points);
                    hDistance[newh] = whoo.distance();
                    hPointNew = whoo.rawPoint();

                    if(hDistance[newh] < hDistance[oldh])
                    {
                        h = hDistance[newh];
                        hPoint = hPointNew;
                        hDistance[oldh] = hDistance[newh];
                        hPointOld = hPointNew;
                        //Sf = td.spc().pMesh().faceAreas()[faceI];

                        Sf = position() - whoo.rawPoint();
                        Sf /= mag(Sf);
                    }
                }
            }
        }
        h -= d_/2.;

        scalar vdwWall = 0.;
        scalar edlWall = 0.;
        short vdwFlag = 0;
        short edlFlag = 0;
        if (td.spc().vdwModel() == "Gregory") {vdwFlag = 1;}
        if (td.spc().vdwModel() == "SchenkelKitchener") {vdwFlag = 2;}
        if (td.spc().vdwModel() == "Czarnecky") {vdwFlag = 3;}
        if (td.spc().edlModel() == "constantCharge") {edlFlag = 1;}
        if (td.spc().edlModel() == "constantPotential") {edlFlag = 2;}
        if (td.spc().edlModel() == "LSA") {edlFlag = 3;}
        
        // some physical constants
        scalar lambda = 1.e-7; //London wavelength
        scalar BoltzConst = 1.38E-23; //Boltzmann constant ((m^2*kg)/(s^2*K))
        scalar eCharge = 1.60E-19; //electron charge (C)
        scalar Avogadro = 6.02214179E23; // Avogadro constant (1/mol)
        scalar permFree = 8.8542E-12;  //permittivity of free space (F/m)
        scalar z = 1.; //valancy of counterions

        scalar Flimit = 1.e-10; //limit imposed on colloidal forces

        //calculate DLVO wall forces
        if (h < 10.*d_ && h > 0.)
        {
           //Info << "Calculating DLVO forces!" << endl;
            switch (vdwFlag){
              case 1:
		//if (h < 0.05*d_)
                //{
                  //Gregory
                  vdwWall = -td.spc().Awall()*d_/12./h/h/(5.32*h/lambda + 1.);
                //  Info << "Using Gregory model for VDW" << endl;
                //}
                //else
                //{
                  //Czarnecki
                //  vdwWall = -td.spc().Awall()*(
                //            2.45*lambda/60/pi*((d_ - h)/pow(h,3) + (2.*d_ + h)/pow((h + d_),3)) -
                //            2.17*pow(lambda,2)/720./pow(pi,2)*((3*d_ - 2.*h)/pow(h,4) + (5.*d_ + 2.*h)/pow((h + d_),4)) +
                //            0.59*pow(lambda,3)/5040./pow(pi,3)*((6.*d_ - 3.*h)/pow(h,5) + (9.*d_ +3.*h)/pow((h + d_),5))
                //            );
                //  Info << "Using Czarnecki model for VDW" << endl;
                //}
                break;
              case 2:
		//if (h < 0.05*d_)
                //{
                  //Schenkel and Kitchener
                  vdwWall = -td.spc().Awall()*d_/12./h/h*(1./(14.*h/lambda + 1.) + 14.*h/lambda/pow((14.*h/lambda + 1.),2));
                //  Info << "Using Schenkel/Kitchener model for VDW" << endl;
                //}
                //else
                //{
                  //Czarnecki
                //  vdwWall = -td.spc().Awall()*(
                //            2.45*lambda/60/pi*((d_ - h)/pow(h,3) + (2.*d_ + h)/pow((h + d_),3)) -
                //            2.17*pow(lambda,2)/720./pow(pi,2)*((3*d_ - 2.*h)/pow(h,4) + (5.*d_ + 2.*h)/pow((h + d_),4)) +
                //            0.59*pow(lambda,3)/5040./pow(pi,3)*((6.*d_ - 3.*h)/pow(h,5) + (9.*d_ +3.*h)/pow((h + d_),5))
                //            );
                //  Info << "Using Czarnecki model for VDW" << endl;
                //}
                break;
	      case 3:
		vdwWall = -td.spc().Awall()*(
                            2.45*lambda/60/pi*((d_ - h)/pow(h,3) + (2.*d_ + h)/pow((h + d_),3)) -
                            2.17*pow(lambda,2)/720./pow(pi,2)*((3*d_ - 2.*h)/pow(h,4) + (5.*d_ + 2.*h)/pow((h + d_),4)) +
                            0.59*pow(lambda,3)/5040./pow(pi,3)*((6.*d_ - 3.*h)/pow(h,5) + (9.*d_ +3.*h)/pow((h + d_),5))
                            );
		break;  
              //default:
                //vdwWall = 0.;
            }

            //Unni and Yang
            //vdwWall = -td.spc().Awall()*lambda*d_*(lambda + 22.22*h)/12./h/h/pow(lambda + 11.11*h,2);

            scalar nIons = td.spc().molarity()*Avogadro/0.001; //number of ions per unit volume (1/m^3)
            scalar eps = permFree*td.spc().perm(); //fluid permittivity
            scalar kappa = sqrt(2.*eCharge*eCharge*nIons*z*z/eps/BoltzConst/td.spc().T()); //Debye-Huckel reciprocal length parameter
            scalar y1 = z*eCharge*td.spc().psiPart()/BoltzConst/td.spc().T(); //Particle reduced surface potential
            scalar y2 = z*eCharge*td.spc().psiWall()/BoltzConst/td.spc().T(); //Wall reduced surface potential
            scalar gamma1 = tanh(y1/4.);
            scalar gamma2 = tanh(y2/4.);

            switch (edlFlag){
              case 1://Constant charge
                edlWall = pi*d_*nIons*BoltzConst*td.spc().T()/kappa*((y1*y1 + y2*y2)*(cosh(kappa*h)/sinh(kappa*h) - 1.)
                          + 2.*y1*y2/sinh(kappa*h));
                //Info << "Using constant charge model for EDL" << endl;
		break;

              case 2://Constant potential
                edlWall = pi*d_*nIons*BoltzConst*td.spc().T()/kappa*((y1*y1 + y2*y2)*(1. - cosh(kappa*h)/sinh(kappa*h))
                          + 2.*y1*y2/sinh(kappa*h));
                //Info << "Using constant potential model for EDL" << endl;
		break;

              case 3://LSA
                edlWall = 64.*pi*d_*gamma1*nIons*BoltzConst*td.spc().T()*gamma2*exp(-kappa*h)/kappa;
                //Info << "Using LSA model for EDL" << endl;
		break;
            }

        }

        //calculate total DLVO force on particle and adjust for maximum value
        vector Fcolloid = vdwWall*Sf + edlWall*Sf;
        if(mag(Fcolloid) > Flimit)
        {
            Fcolloid /= mag(Fcolloid);
            Fcolloid *= Flimit;
        }
        Ucolloid = Fcolloid/m_;

        //Info << "h = " << h << ", Fvdw = " << vdwWall << ", Fedl = " << edlWall
        //     << ", Fcolloid = " << Fcolloid << ", Fgrav = " << (1.0 - rhoc/rhop)*td.g()*m_ << endl;

        U_ = (U_ + dt*(Ucolloid + Dc*Uc + (1.0 - rhoc/rhop)*td.g()))/(1.0 + dt*Dc);

        td.spc().smoment()[celli] += m_*(U_ - Uc)*Dc*dt;

        if (onBoundary() && td.keepParticle)
        {
            if (face() > -1)
            {
                if (isType<processorPolyPatch>(pbMesh[patch(face())]))
                {
                    td.switchProcessor = true;
                }
            }
        }
    }

    return td.keepParticle;
}


void Foam::colloidParticle::hitProcessorPatch
(
    const processorPolyPatch&,
    colloidParticle::trackData& td
)
{
    td.switchProcessor = true;
    //Info << "Switching processor patch!" << endl;
}


void Foam::colloidParticle::hitProcessorPatch
(
    const processorPolyPatch&,
    int&
)
{}


void Foam::colloidParticle::hitWallPatch
(
    const wallPolyPatch& wpp,
    colloidParticle::trackData& td
)
{
    vector nw = wpp.faceAreas()[wpp.whichFace(face())];
    nw /= mag(nw);

    scalar Un = U_ & nw;
    vector Ut = U_ - Un*nw;

    if (Un > 0)
    {
        U_ -= (1.0 + td.spc().e())*Un*nw;
    }

    U_ -= td.spc().mu()*Ut;

    //Info << "COLLISION WITH WALL!!!!" << endl;
}


void Foam::colloidParticle::hitWallPatch
(
    const wallPolyPatch&,
    int&
)
{}


void Foam::colloidParticle::hitPatch
(
    const polyPatch&,
    colloidParticle::trackData& td
)
{
    td.keepParticle = false;
    //Info << "Hitting patch, particle removed!" << endl;
}


void Foam::colloidParticle::hitPatch
(
    const polyPatch&,
    int&
)
{}


void Foam::colloidParticle::transformProperties (const tensor& T)
{
    Particle<colloidParticle>::transformProperties(T);
    U_ = transform(T, U_);
    //Info << "Particle transformProperties?" << endl;
}


void Foam::colloidParticle::transformProperties(const vector& separation)
{
    Particle<colloidParticle>::transformProperties(separation);
}


// ************************************************************************* //
