/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2008 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

\*---------------------------------------------------------------------------*/

#include "colloidParticleCloud.H"
#include "collisionModel.H"
#include "fvMesh.H"
#include "volFields.H"
#include "interpolationCellPoint.H"
#include "volPointInterpolation.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineParticleTypeNameAndDebug(colloidParticle, 0);
    defineTemplateTypeNameAndDebug(Cloud<colloidParticle>, 0);
};

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::colloidParticleCloud::colloidParticleCloud
(
    const fvMesh& mesh,
    const word& cloudName
)
:
    Cloud<colloidParticle>(mesh, cloudName, false),
    mesh_(mesh),
    runTime_(mesh.time()),
    random_(666),
    particleProperties_
    (
        IOobject
        (
            "particleProperties",
            mesh_.time().constant(),
            mesh_,
            IOobject::MUST_READ,
           IOobject::NO_WRITE
        )
    ),
    transportProperties_
    (
        IOobject
        (
            "transportProperties",
            mesh_.time().constant(),
            mesh_,
            IOobject::MUST_READ,
            IOobject::NO_WRITE
        )
    ),

    collisionModel_
    (
        collisionModel::New
        (
            particleProperties_,
            *this,
            random_
        )
    ),

    rhop_(dimensionedScalar(particleProperties_.lookup("rhop")).value()),
    e_(dimensionedScalar(particleProperties_.lookup("e")).value()),
    mu_(dimensionedScalar(particleProperties_.lookup("mu")).value()),

    useSourceMoment_(readBool(particleProperties_.lookup("useSourceMoment"))),
    inject_(readBool(particleProperties_.lookup("inject"))),
    vdwModel_(word(particleProperties_.lookup("vdwModel"))),
    edlModel_(word(particleProperties_.lookup("edlModel"))),
		
    ppts0_(readScalar(particleProperties_.subDict("injection").lookup("ppts0"))),
    ppts1_(readScalar(particleProperties_.subDict("injection").lookup("ppts1"))),
    c0_(particleProperties_.subDict("injection").lookup("c0")),
    c1_(particleProperties_.subDict("injection").lookup("c1")),
    vel0_(readScalar(particleProperties_.subDict("injection").lookup("vel0"))),
    vel1_(particleProperties_.subDict("injection").lookup("vel1")),
    d0_(readScalar(particleProperties_.subDict("injection").lookup("d0"))),
    d1_(readScalar(particleProperties_.subDict("injection").lookup("d1"))),
    tStart_(readScalar(particleProperties_.subDict("injection").lookup("tStart"))),
    tEnd_(readScalar(particleProperties_.subDict("injection").lookup("tEnd"))),

    Awall_(readScalar(particleProperties_.subDict("colloidal").lookup("Awall"))),
    Apart_(readScalar(particleProperties_.subDict("colloidal").lookup("Apart"))),
    psiWall_(readScalar(particleProperties_.subDict("colloidal").lookup("psiWall"))),
    psiPart_(readScalar(particleProperties_.subDict("colloidal").lookup("psiPart"))),
    molarity_(readScalar(particleProperties_.subDict("colloidal").lookup("molarity"))),
    perm_(readScalar(particleProperties_.subDict("colloidal").lookup("perm"))),
    T_(readScalar(particleProperties_.subDict("colloidal").lookup("T"))),

    rhoc_(dimensionedScalar(transportProperties_.lookup("rho")).value()),
    nuc_(dimensionedScalar(transportProperties_.lookup("nu")).value()),

    smoment_(mesh.nCells(), vector::zero)
{
    colloidParticle::readFields(*this);
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::colloidParticleCloud::move(const dimensionedVector& g)
{
    smoment_ = vector::zero;

    //const volScalarField& rho = mesh_.lookupObject<const volScalarField>("rho");
    const volVectorField& U = mesh_.lookupObject<const volVectorField>("U");
    //const volScalarField& nu = mesh_.lookupObject<const volScalarField>("nu");

    pointMesh pMesh(U.mesh());
    volPointInterpolation vpi(U.mesh(), pMesh);

    //interpolationCellPoint<scalar> rhoInterp(vpi, rho);
    interpolationCellPoint<vector> UInterp(vpi, U);
    //interpolationCellPoint<scalar> nuInterp(vpi, nu);

    colloidParticle::trackData td(*this, UInterp, g.value());

    if(inject_) {
        scalar ppts = random().GaussNormal()*ppts1_ + ppts0_;
        Info << "Injecting " << ppts << " particles." << endl;
        scalar injects;
        for (injects = 0; injects < ppts; injects++)
            this->inject(td);
    }

    label thisMany_ = Cloud<colloidParticle>::size();
    Info << thisMany_ << " particles are now in the domain." << endl;

    Info << "Checkpoint A" << endl;
    collisions().collideParticles(runTime_.deltaT().value());

    Info << "Checkpoint B" << endl;
    Cloud<colloidParticle>::move(td);

    Info << "Checkpoint C" << endl;
}

void Foam::colloidParticleCloud::inject(colloidParticle::trackData &td) {
    if(mesh().time().value()<tStart_ || mesh().time().value()>tEnd_) {
        return;
    }

    scalar dx = c1_.x() - c0_.x();
    scalar dy = c1_.y() - c0_.y();
    scalar dz = c1_.z() - c0_.z();
    scalar xvar = random().scalar01();
    scalar yvar = random().scalar01();
    scalar zvar = random().scalar01();
    vector tmp = vector(xvar*dx,yvar*dy,zvar*dz);
    vector pos = c0_ + tmp;

    vector vel=random().vector01()*vel0_+vel1_;

    scalar d=fabs(random().GaussNormal())*d1_+d0_;

    label cellI=mesh_.findCell(pos);

    if(cellI>=0) {
        colloidParticle* ptr=new colloidParticle(*this,pos,cellI,d,vel);

        ptr->stepFraction() = 1;

        addParticle(ptr);

        //td.countInject();
    }
}

void Foam::colloidParticleCloud::writeFields() const
{
    colloidParticle::writeFields(*this);
}


// ************************************************************************* //
