/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2005 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    

\*---------------------------------------------------------------------------*/

#include "error.H"

#include "HSModel.H"
#include "addToRunTimeSelectionTable.H"
#include <vector>
#include <algorithm>
#include <list>

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

defineTypeNameAndDebug(HSModel, 0);

addToRunTimeSelectionTable
(
    collisionModel,
    HSModel,
    dictionary
);


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

// Construct from components
HSModel::HSModel
(
    const dictionary& dict,
    colloidParticleCloud& sm,
    Random& rndGen
)
:
    collisionModel(dict, sm, rndGen),
    vols_(sm.mesh().V())
 
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

HSModel::~HSModel()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void HSModel::collideParticles(const scalar dt) const
{

  std::vector<colloidParticle*>  cellParts[InCloud_.mesh().cells().size()]; 
  scalar VR=0.; // relative speed of two selected particles
  vector VCCM,VRCP;
  label tmp[50]; //Assumes no more than 50 particles/cell
  label count=0;
  vector VRC;

  tmp[0]=0;


  for(colloidParticleCloud::iterator elmnt = InCloud_.begin(); elmnt != InCloud_.end(); ++elmnt)
    {
      count++;
      label celli = elmnt().cell();
      tmp[count]=celli;
        
      if(tmp[count]==tmp[count-1]) //put particles into vector list
	{
	  colloidParticle* ptr = &(elmnt());
	  cellParts[celli].push_back(ptr);
	}
      else
	{
	  colloidParticle* ptr = &(elmnt());
	  cellParts[celli].push_back(ptr);
	  count = 0;
	}

    }  //end loop through the elements in the cloud 

  forAll(InCloud_.mesh().cells(),i) //consider collisions in cell i
    {
      scalar N = cellParts[i].size();

      if( N > 1) //if there exist more than one particle in cell i
	{
	  for(label k=0; k < N; k++) //loop through the particles
	    {
	      vector d1 = cellParts[i][k]->position();
	      vector d2 = cellParts[i][k+1]->position();
              vector drel = d2 - d1;

	      scalar dmag = mag(drel);

	      scalar dia1 = cellParts[i][k]->d();
	      scalar dia2 = cellParts[i][k+1]->d();

	      if(dmag <= 0.5*(dia1 + dia2)) //within collision range per HS CCS;
		{

		  //compute the relative velocity and its magnitude
		  VRC = cellParts[i][k]->U() - cellParts[i][k+1]->U(); 
		  VR = mag(VRC);

                  scalar approach = VRC.x()*drel.x() + VRC.y()*drel.y() + VRC.z()*drel.z();
          
                  if(approach > 0.) //particles still approaching each other; implementing elastic collisions
                    {
                      Info << "Particles colliding!" << endl;
		      scalar SPM = (cellParts[i][k]->m()/(cellParts[i][k]->m() + cellParts[i][k+1]->m())) *
		        cellParts[i][k+1]->m();
                 
		      scalar RML = SPM / cellParts[i][k+1]->m();
		      scalar RMM = SPM / cellParts[i][k]->m(); 
                                        
		      //VMMC is the center of mass velocity
		      VCCM.x() = RML * cellParts[i][k]->U().x() + RMM * cellParts[i][k+1]->U().x();
		      VCCM.y() = RML * cellParts[i][k]->U().y() + RMM * cellParts[i][k+1]->U().y();
		      VCCM.z() = RML * cellParts[i][k]->U().z() + RMM * cellParts[i][k+1]->U().z();

		      scalar B = 2. * rndGen_.scalar01() -1.;  //B is the cosine of a random elevation angle
		      scalar A = sqrt(1. - B * B);

		      VRCP.x() = B * VR;
		      scalar C = 2. * mathematicalConstant::pi * rndGen_.scalar01();  //C is a random azimuth angle
		      VRCP.y() = A * cos(C) * VR;
		      VRCP.z() = A * sin(C) * VR;

		      cellParts[i][k]->setU().x() = VCCM.x() + VRCP.x() * RMM;
		      cellParts[i][k]->setU().y() = VCCM.y() + VRCP.y() * RMM;
		      cellParts[i][k]->setU().z() = VCCM.z() + VRCP.z() * RMM;

		      cellParts[i][k+1]->setU().x() = VCCM.x() - VRCP.x() * RML;
		      cellParts[i][k+1]->setU().y() = VCCM.y() - VRCP.y() * RML;
		      cellParts[i][k+1]->setU().z() = VCCM.z() - VRCP.z() * RML;

                    }
                               
		}
               
                
	      if((k+1) == (N-1)) k++;
		 
                      
	    } //end for(label k...
	}//end if (N > 1)...

         

    }	 //end cells 


}    

         


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
